﻿using System;
using System.Collections.Generic;

namespace ELTE.Basics.Polynomial
{
    /// <summary>
    /// Polinom típusa.
    /// </summary>
    class Polynomial
    {
        private Double[] _Coefficients; // együtthatókat tároló tömb

        /// <summary>
        /// Polinom méretének lekérdezése.
        /// </summary>
        public Int32 Size { get { return _Coefficients.Length; } }
        /// <summary>
        /// Polinom fokszámának lekérdezése.
        /// </summary>
        public Int32 Degree { get { return _Coefficients.Length - 1; } }

        /// <summary>
        /// Polinom együtthatójának lekérdezése vagy beállítása.
        /// </summary>
        /// <param name="degree">Az együttható fokszáma.</param>
        /// <returns>Az adott fokszámú együttható.</returns>
        public Double this[Int32 degree] // indexelő operátor, ugyanúgy használjuk, mint egy tulajdonságot
        {
            get 
            {
                return Get(degree); // olvasás, tovább dobhat kivételt
            }
            set
            {
                Set(degree, value); // beállítása, tovább dobhat kivételt
            }
        }

        /// <summary>
        /// Polinom létrehozása.
        /// </summary>
        public Polynomial()
        {
            _Coefficients = new Double[0]; // üres polinom létehozása
        }

        /// <summary>
        /// Polinom létrehozása.
        /// </summary>
        /// <param name="degree">A polinom fokszáma.</param>
        public Polynomial(Int32 degree)
        {
            if (degree >= 0) // ha a méret pozitív
            {
                _Coefficients = new Double[degree + 1]; // létrehozzuk a megfelelő méretű tömböt
                for (Int32 i = 0; i < _Coefficients.Length; i++)
                    _Coefficients[i] = 1; // minden együttható 1 lesz
            }
            else
            {
                throw new ArgumentException(); // jelezzük, hogy hiba van a metódus argumentumaiban
            }
        }

        /// <summary>
        /// Polinom létrehozása.
        /// </summary>
        /// <param name="coefficients">A polinom együtthatói.</param>
        public Polynomial(Double[] coefficients)
        {
            _Coefficients = new Double[coefficients.Length];
            for (Int32 i = 0; i < _Coefficients.Length; i++)
                _Coefficients[i] = coefficients[i];
        }

        /// <summary>
        /// Polinom együtthatójának lekérdezése.
        /// </summary>
        /// <param name="degree">Az együttható fokszáma.</param>
        /// <returns>Az adott fokszámú együttható.</returns>
        public Double Get(Int32 degree) // adott fokszámú együttható lekérdezése
        {
            if (degree >= 0 && degree < Size) // ha méreten belül van a fokszám
            {
                return _Coefficients[degree]; // visszaadjuk a megfelelő értéket
            }
            else
            {
                throw new IndexOutOfRangeException(); // jelezzük, hogy hiba van a megadott fokszámban
            }
        }

        /// <summary>
        /// Polinom együtthatójának beállítása.
        /// </summary>
        /// <param name="degree">Az együttható fokszáma.</param>
        /// <param name="coefficient">Az együttható értéke.</param>
        public void Set(Int32 degree, Double coefficient) // adott fokszámú együttható beállítása
        {
            if (degree >= 0 && degree < Size) // ha méreten belül van a fokszám
            {
                _Coefficients[degree] = coefficient; // beállítjuk a megfelelő értéket
            }
            else if (degree >= 0) // ha pozitív a fokszám, de a méreten kívül van
            {
                Double[] newCoeffitients = new Double[degree + 1]; // új tömböt kell létrehoznunk
                for (Int32 i = 0; i < _Coefficients.Length; i++) // bemásoljuk az értékeket a régiből
                {
                    newCoeffitients[i] = _Coefficients[i];
                }
                newCoeffitients[degree] = coefficient;
                _Coefficients = newCoeffitients; // émajd kicseréljük a két tömböt
            }
            else
            {
                throw new ArgumentException(); // jelezzük, hogy hiba van a metódus argumentumaiban
            }
        }

        /// <summary>
        /// A polinom kirtékelése.
        /// </summary>
        /// <param name="value">A kiértékelt pozíció.</param>
        /// <returns>A kiértékelt polinom.</returns>
        public Double Evaluate(Double value)
        {
            Double result = 0; // kiértékelés Horner-algoritmussal
            for (Int32 i = _Coefficients.Length - 1; i >= 0; i--)
            {
                result = result * value + _Coefficients[i]; 
            }
            return result;
        }

        /// <summary>
        /// Polinomok összeadása.
        /// </summary>
        /// <param name="first">Balérték.</param>
        /// <param name="second">Jobbérték.</param>
        /// <returns>Az eredmény polinom.</returns>
        public static Polynomial operator +(Polynomial first, Polynomial second)
        {
            Polynomial result = new Polynomial((first.Size > second.Size) ? first.Size : second.Size); 
            // amelyik a nagyobb, ahhoz kell igazítani a méretet
            for (Int32 i = 0; i < first.Size || i < second.Size; i++)
            {
                if (i < first.Size && i < second.Size) // ha mindekttőben van az adott tag
                    result._Coefficients[i] = first._Coefficients[i] + second._Coefficients[i];
                else if (i < first.Size) // ha csak egyikben
                    result._Coefficients[i] = first._Coefficients[i];
                else if (i < second.Size)
                    result._Coefficients[i] = second._Coefficients[i];
            }
            return result;
        }
    }
}
